/* -*-c-*-
 *
 * $Id: conf.h,v 1.2 1999/07/26 23:28:39 mdw Exp $
 *
 * Configuration parsing
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the `fw' port forwarder.
 *
 * `fw' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * `fw' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with `fw'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: conf.h,v $
 * Revision 1.2  1999/07/26 23:28:39  mdw
 * Major reconstruction work for new design.
 *
 * Revision 1.1.1.1  1999/07/01 08:56:23  mdw
 * Initial revision.
 *
 */

#ifndef CONF_H
#define CONF_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <mLib/dstr.h>

#ifndef SCAN_H
#  include "scan.h"
#endif

/*----- Magic numbers -----------------------------------------------------*/

#define CTOK_EOF (-1)
#define CTOK_WORD 256

/*----- Functions provided ------------------------------------------------*/

/* --- @token@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to scanner definition
 *
 * Returns:	Type of token scanned.
 *
 * Use:		Reads the next token from the character scanner.
 */

extern int token(scanner */*sc*/);

/* --- @error@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to scanner definition
 *		@const char *msg@ = message skeleton string
 *		@...@ = extra arguments for the skeleton
 *
 * Returns:	Doesn't
 *
 * Use:		Reports an error at the current scanner location.
 */

extern void error(scanner */*sc*/, const char */*msg*/, ...);

/* --- @conf_enum@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to a scanner object
 *		@const char *list@ = comma-separated things to allow
 *		@unsigned @f = flags for the search
 *		@const char *err@ = error message if not found
 *
 * Returns:	Index into list, zero-based, or @-1@.
 *
 * Use:		Checks whether the current token is a string which matches
 *		one of the comma-separated items given.  If not, an error is
 *		reported; otherwise the index of the matched item is
 *		returned.
 */

#define ENUM_ABBREV 1u
#define ENUM_NONE 2u

extern int conf_enum(scanner */*sc*/, const char */*list*/,
		     unsigned /*flags*/, const char */*err*/);

/* --- @conf_prefix@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to a scanner object
 *		@const char *p@ = pointer to prefix string to check
 *
 * Returns:	Nonzero if the prefix matches.
 *
 * Use:		If the current token is a word matching the given prefix
 *		string, then it and an optional `.' character are removed and
 *		a nonzero result is returned.  Otherwise the current token is
 *		left as it is, and zero is returned.
 *
 *		Typical options parsing code would remove an expected prefix,
 *		scan an option anyway (since qualifying prefixes are
 *		optional) and if a match is found, claim the option.  If no
 *		match is found, and a prefix was stripped, then an error
 *		should be reported.
 */

extern int conf_prefix(scanner */*sc*/, const char */*p*/);

/* --- @CONF_BEGIN@, @CONF_ACCEPT@, @CONF_END@ --- *
 *
 * Arguments:	@sc@ = scanner to read from
 *		@prefix@ = prefix to scan for
 *		@desc@ = description of what we're parsing
 *
 * Use:		Bracket an options parsing routine.
 */

#define CS_PLAIN 0
#define CS_PREFIX 1
#define CS_BRACE 2
#define CS_UNKNOWN 3

#define CONF_BEGIN(sc, prefix, desc) do {				\
  scanner *_conf_sc = (sc);						\
  const char *_conf_desc = (desc);					\
  int _conf_state = CS_PLAIN;						\
  if (_conf_sc->t == CTOK_WORD &&					\
      strcmp(_conf_sc->d.buf, (prefix)) == 0) {				\
    token(_conf_sc);							\
    _conf_state = CS_PREFIX;						\
    if (_conf_sc->t == '.')						\
      token(_conf_sc);							\
    else if (_conf_sc->t == '{') {					\
      token(_conf_sc);							\
      _conf_state = CS_BRACE;						\
    }									\
  }									\
  if (_conf_sc->t != CTOK_WORD)						\
    error(_conf_sc, "parse error, expected option keyword");		\
  do {

#define CONF_ACCEPT goto _conf_accept
#define CONF_REJECT goto _conf_reject
#define CONF_QUAL (_conf_state != CS_PLAIN)

#define CONF_END							\
  _conf_reject:								\
    if (_conf_state == CS_PLAIN)					\
      _conf_state = CS_UNKNOWN;						\
    else {								\
      error(_conf_sc, "unknown %s option `%s'",				\
	    _conf_desc, _conf_sc->d.buf);				\
    }									\
  _conf_accept:								\
    if (_conf_state == CS_BRACE && _conf_sc->t == ';')			\
      token(_conf_sc);							\
  } while (_conf_state == CS_BRACE && _conf_sc->t == CTOK_WORD);	\
  if (_conf_state == CS_BRACE) {					\
    if (_conf_sc->t == '}')						\
      token(_conf_sc);							\
    else								\
      error(_conf_sc, "parse error, expected `}'");			\
  }									\
  return (_conf_state != CS_UNKNOWN);					\
} while (0)

/* --- @conf_name@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to scanner
 *		@char delim@ = delimiter character to look for
 *		@dstr *d@ = pointer to dynamic string for output
 *
 * Returns:	---
 *
 * Use:		Reads in a compound name consisting of words separated by
 *		delimiters.  Leading and trailing delimiters are permitted,
 *		although they'll probably cause confusion if used.  The name
 *		may be enclosed in square brackets if that helps at all.
 *
 *		Examples of compound names are filenames (delimited by `/')
 *		and IP addresses (delimited by `.').
 */

extern void conf_name(scanner */*sc*/, char /*delim*/, dstr */*d*/);

/* --- @conf_parse@ --- *
 *
 * Arguments:	@scanner *sc@ = pointer to a scanner structure
 *
 * Returns:	---
 *
 * Use:		Parses a configuration file fragment from the scanner
 */

extern void conf_parse(scanner *sc);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
